/**
 * ASTStatementList.java
 * 
 * Copyright (c) 2006 David Holroyd
 */

package uk.co.badgersinfoil.metaas.impl;

import org.antlr.runtime.RecognitionException;
import org.asdt.core.internal.antlr.AS3Parser;
import uk.co.badgersinfoil.metaas.SyntaxException;
import uk.co.badgersinfoil.metaas.dom.ASBlock;
import uk.co.badgersinfoil.metaas.dom.ASIfStatement;
import uk.co.badgersinfoil.metaas.dom.ASSwitchStatement;
import uk.co.badgersinfoil.metaas.dom.StatementContainer;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListToken;
import uk.co.badgersinfoil.metaas.impl.antlr.LinkedListTree;

/**
 * Implements the behavior of StatementContainer.  May be managing statements
 * for either be a block, or a switch-statement label.
 */
public class ASTStatementList implements ASBlock {

	private LinkedListTree ast;

	public ASTStatementList(LinkedListTree ast) {
		this.ast = ast;
	}

	public void addStmt(String statement) {
		// We need to add a ';' so that LA(2) at the end of an
		// expression-statement will not unexpectedly find EOF. we add
		// a second ';' so that the trailing ';' can simply be ommitted
		// by the caller
		// TODO: assert that there really is no extra code trailing the
		//       parsed statement
		AS3Parser parser = ASTUtils.parse(statement+";;");
		LinkedListTree stmt;
		try {
			stmt = ASTUtils.tree(parser.statement());
		} catch (RecognitionException e) {
			throw ASTUtils.buildSyntaxException(statement, parser, e);
		}
		stmt.getStopToken().setNext(null);
		addStatement(stmt);
	}

	public void addComment(String text) {
		LinkedListToken comment = TokenBuilder.newSLComment("//" + text);
		String indent = findIndentForComment();
		ast.appendToken(TokenBuilder.newNewline());
		ast.appendToken(TokenBuilder.newWhiteSpace(indent));
		ast.appendToken(comment);
	}

	private String findIndentForComment() {
		LinkedListTree last = ast.getLastChild();
		String indent;
		if (last == null) {
			indent = "\t" + ASTUtils.findIndent(ast);
		} else {
			indent = ASTUtils.findIndent(last);
		}
		return indent;
	}

	public ASIfStatement newIf(String condition) {
		LinkedListTree ifStmt = ASTUtils.newAST(AS3Parser.IF, "if");
		ifStmt.appendToken(TokenBuilder.newSpace());
		ifStmt.addChildWithTokens(ASTUtils.parseCondition(condition));
		ifStmt.appendToken(TokenBuilder.newSpace());
		ifStmt.addChildWithTokens(ASTFactory.newBlock());
		addStatement(ifStmt);
		return new ASTASIfStatement(ifStmt);
	}

	private LinkedListTree parseForInit(String expr) {
		AS3Parser parser = ASTUtils.parse(expr + ";");
		try {
			return ASTUtils.tree(parser.forInit());
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	private LinkedListTree parseForCond(String expr) {
		AS3Parser parser = ASTUtils.parse(expr + ";");
		try {
			return ASTUtils.tree(parser.forCond());
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	private LinkedListTree parseForIter(String expr) {
		AS3Parser parser = ASTUtils.parse(expr + ")");
		try {
			return ASTUtils.tree(parser.forIter());
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	public StatementContainer newFor(String init, String condition, String iterate) {
		LinkedListTree forStmt = ASTUtils.newAST(AS3Parser.FOR, "for");
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.appendToken(TokenBuilder.newLParen());
		if (init != null) {
			forStmt.addChildWithTokens(parseForInit(init));
		} else {
			LinkedListTree initStmt = ASTUtils.newAST(AS3Parser.FOR_INIT);
			forStmt.addChildWithTokens(initStmt);
		}
		forStmt.appendToken(TokenBuilder.newSemi());
		forStmt.appendToken(TokenBuilder.newSpace());
		if (condition != null) {
			forStmt.addChildWithTokens(parseForCond(condition));
		} else {
			LinkedListTree condStmt = ASTUtils.newAST(AS3Parser.FOR_CONDITION);
			forStmt.addChildWithTokens(condStmt);
		}
		forStmt.appendToken(TokenBuilder.newSemi());
		forStmt.appendToken(TokenBuilder.newSpace());
		if (iterate != null) {
			forStmt.addChildWithTokens(parseForIter(iterate));
		} else {
			LinkedListTree iterStmt = ASTUtils.newAST(AS3Parser.FOR_ITERATOR);
			forStmt.addChildWithTokens(iterStmt);
		}
		forStmt.appendToken(TokenBuilder.newRParen());
		forStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTFactory.newBlock();
		forStmt.addChildWithTokens(block);
		addStatement(forStmt);
		return new ASTStatementList(block);
	}

	private LinkedListTree parseDeclaration(String decl) {
		AS3Parser parser = ASTUtils.parse(decl + ";");
		try {
			return ASTUtils.tree(parser.declaration());
		} catch (RecognitionException e) {
			throw new SyntaxException(e);
		}
	}

	public StatementContainer newForIn(String declaration, String expression) {
		LinkedListTree forStmt = ASTUtils.newAST(AS3Parser.FOR, "for");
		forStmt.appendToken(TokenBuilder.newSpace());
		genForInSetup(forStmt, declaration, expression);
		forStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTFactory.newBlock();
		forStmt.addChildWithTokens(block);
		addStatement(forStmt);
		return new ASTStatementList(block);
	}

	public StatementContainer newForEachIn(String declaration, String expression) {
		LinkedListTree forStmt = ASTUtils.newAST(AS3Parser.FOR_EACH, "for");
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.appendToken(TokenBuilder.newEach());
		genForInSetup(forStmt, declaration, expression);
		forStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTFactory.newBlock();
		forStmt.addChildWithTokens(block);
		addStatement(forStmt);
		return new ASTStatementList(block);
	}

	/**
	 * Common code for both for-in and for-each-in loop setup
	 */
	private void genForInSetup(LinkedListTree forStmt, String declaration, String expression) {
		forStmt.appendToken(TokenBuilder.newLParen());
		forStmt.addChildWithTokens(parseDeclaration(declaration));
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.appendToken(TokenBuilder.newIn());
		forStmt.appendToken(TokenBuilder.newSpace());
		forStmt.addChildWithTokens(ASTUtils.parseExpr(expression));
		forStmt.appendToken(TokenBuilder.newRParen());
	}

	public StatementContainer newWhile(String condition) {
		LinkedListTree whileStmt = ASTUtils.newAST(AS3Parser.WHILE, "while");
		whileStmt.appendToken(TokenBuilder.newSpace());
		whileStmt.addChildWithTokens(ASTUtils.parseCondition(condition));
		whileStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTFactory.newBlock();
		whileStmt.addChildWithTokens(block);
		addStatement(whileStmt);
		return new ASTStatementList(block);
	}

	public StatementContainer newDoWhile(String condition) {
		LinkedListTree doWhileStmt = ASTUtils.newAST(AS3Parser.DO, "do");
		doWhileStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTFactory.newBlock();
		doWhileStmt.addChildWithTokens(block);
		doWhileStmt.appendToken(TokenBuilder.newSpace());
		doWhileStmt.appendToken(TokenBuilder.newWhile());
		doWhileStmt.appendToken(TokenBuilder.newSpace());
		doWhileStmt.addChildWithTokens(ASTUtils.parseCondition(condition));
		doWhileStmt.appendToken(TokenBuilder.newSemi());
		addStatement(doWhileStmt);
		return new ASTStatementList(block);
	}

	public ASSwitchStatement newSwitch(String condition) {
		LinkedListTree switchStmt = ASTUtils.newAST(AS3Parser.SWITCH, "switch");
		switchStmt.appendToken(TokenBuilder.newSpace());
		switchStmt.addChildWithTokens(ASTUtils.parseCondition(condition));
		switchStmt.appendToken(TokenBuilder.newSpace());
		LinkedListTree block = ASTFactory.newBlock();
		switchStmt.addChildWithTokens(block);
		addStatement(switchStmt);
		return new ASTASSwitchStatement(block);
	}

	private void addStatement(LinkedListTree stmt) {
		ASTUtils.addChildWithIndentation(ast, stmt);
	}

	public boolean containsCode() {
		return ast.getFirstChild() != null;
	}

	public LinkedListTree getAST() {
		return ast;
	}
}